"""
send message reactions to single contact
"""


from dataclasses import dataclass, field
from typing import List, Optional

from sanic import Blueprint
from sanic.log import logger
from sanic.response import json
from sanic_ext import openapi, validate
from signal_cli_dbus_rest_api.dataclasses import Error, ResponseTimestamp
from signal_cli_dbus_rest_api.lib.dbus import SignalCLIDBus
from signal_cli_dbus_rest_api.lib.helper import (get_groupid_as_bytes,
                                                 is_phone_number)

reactions_v1 = Blueprint("reactions_v1", url_prefix="/reactions")


def send_reaction(
    number: str,
    reaction: str,
    remove: bool,
    recipients: str,
    target_author: str,
    timestamp: int,
):  # pylint: disable=too-many-arguments
    """
    send reaction
    """
    try:
        if not isinstance(recipients, list):
            recipients = [recipients]
        dbus = SignalCLIDBus(number=number)
        real_recipients = is_phone_number(
            recipients=recipients, dbus_connection=dbus.dbusconn
        )
        method = dbus.dbusconn.sendMessageReaction
        signature = "sbsxas"
        if not real_recipients:
            real_recipients = get_groupid_as_bytes(recipient=recipients[0])
            method = dbus.dbusconn.sendGroupMessageReaction
            signature = "sbsxay"
        if not real_recipients:
            return (
                {
                    "error": f"{recipients} is neither a phone number nor a valid group id"
                },
                400,
            )
        result = method(
            reaction,
            remove,
            target_author,
            timestamp,
            real_recipients,
            signature=signature,
        )
        return {"timestamp": str(result)}, 200
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return {"error": error}, 400


@dataclass
class ReactionsV1PostParams:
    """
    ReactionsV1PostParams
    """

    reaction: str
    target_author: str
    timestamp: int
    recipients: Optional[List[str]] = field(default_factory=list)
    recipient: Optional[str] = field(default_factory=str)
    remove: Optional[bool] = False


@reactions_v1.post("/<number:path>", version=1)
@openapi.tag("Reactions")
@openapi.parameter("number", str, required=True, location="path")
@openapi.body({"application/json": ReactionsV1PostParams}, required=True)
@openapi.response(201, {"application/json": ResponseTimestamp}, description="Created")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description(
    "Send a reaction. Request must include either `recipient` or `recipients`!"
)
@validate(json=ReactionsV1PostParams)
async def reactions_v1_post(
    request, number, body: ReactionsV1PostParams
):  # pylint: disable=unused-argument
    """
    Send a reaction.
    """
    if not body.recipient or body.recipients:
        return json(
            {
                "error": "missing either parameter recipient (str) or recipients (list of strings)"
            },
            400,
        )
    return_message, return_code = send_reaction(
        number=number,
        reaction=body.reaction,
        remove=body.remove,
        target_author=body.target_author,
        timestamp=body.timestamp,
        recipients=body.recipients or body.recipient,
    )
    return json(return_message, return_code)


@dataclass
class ReactionsV1DeleteParams:
    """
    ReactionsV1DeleteParams
    """

    recipient: str
    target_author: str
    timestamp: int


@reactions_v1.delete("/<number:path>", ignore_body=False, version=1)
@openapi.tag("Reactions")
@openapi.parameter("number", str, required=True, location="path")
@openapi.body({"application/json": ReactionsV1DeleteParams}, required=True)
@openapi.response(200, {"application/json": ResponseTimestamp}, description="Deleted")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Delete a reaction.")
@validate(ReactionsV1DeleteParams)
async def reactions_v1_delete(
    request, number, body: ReactionsV1DeleteParams
):  # pylint: disable=unused-argument
    """
    Delete a reaction.
    """
    return_message, return_code = send_reaction(
        number=number,
        reaction="👍",
        remove=True,
        target_author=body.target_author,
        timestamp=body.timestamp,
        recipients=body.recipient,
    )
    return json(return_message, return_code)



"""
common dataclasses
"""

from dataclasses import dataclass


@dataclass
class Error:
    """
    SearchV1Error
    """
    error: str


@dataclass
class ResponseTimestamp:
    """
    ResponseTimestamp
    """

    timestamp: str


@dataclass
class GroupId:
    """
    GroupId
    """

    group_id: str



"""
api blueprints group
"""


from sanic import Blueprint
from signal_cli_dbus_rest_api.api.about import about_v1
from signal_cli_dbus_rest_api.api.groups import (create_group_v1,
                                                 delete_group_v1,
                                                 group_details_v1,
                                                 groups_for_number_v1,
                                                 join_group_v1,
                                                 quit_group_v1,
                                                 update_group_v1)
from signal_cli_dbus_rest_api.api.reactions import reactions_v1
from signal_cli_dbus_rest_api.api.register_verify import register_v1, verify_v1
from signal_cli_dbus_rest_api.api.search import search_v1
from signal_cli_dbus_rest_api.api.send import send_v1, send_v2

entrypoint = Blueprint.group(
    about_v1,
    create_group_v1,
    delete_group_v1,
    groups_for_number_v1,
    group_details_v1,
    join_group_v1,
    quit_group_v1,
    reactions_v1,
    register_v1,
    search_v1,
    send_v1,
    send_v2,
    update_group_v1,
    verify_v1,
    url_prefix="/",
)



"""
register and verify
"""

from dataclasses import dataclass, field
from typing import Optional

from sanic import Blueprint
from sanic.log import logger
from sanic.response import json
from sanic_ext import openapi, validate
from signal_cli_dbus_rest_api.dataclasses import Error
from signal_cli_dbus_rest_api.lib.dbus import SignalCLIDBus

register_v1 = Blueprint("register_v1", url_prefix="/register")
verify_v1 = Blueprint("verify_v1", url_prefix="/register")


@dataclass
class RegisterV1PostParams:
    """
    RegisterV1PostParams
    """

    captcha: Optional[str]
    use_voice: Optional[bool] = False


@register_v1.middleware("request")
async def register_post_dummy_body(request):
    """
    Create empty dummy body for @validate
    if there is none in request
    """
    if not request.body:
        request.body = b"{}"


@register_v1.post("/<number:path>", version=1)
@openapi.tag("Devices")
@openapi.parameter("number", str, required=True, location="path")
@openapi.body({"application/json": RegisterV1PostParams})
@openapi.response(201, {"application/json": None}, description="Created")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Register a phone number with the signal network.")
@validate(RegisterV1PostParams)
async def register_post(
    request, number, body: RegisterV1PostParams
):  # pylint: disable=unused-argument
    """
    Register a phone number.
    """
    opts = [number, body.use_voice]
    try:
        dbus = SignalCLIDBus()
        if body.captcha:
            method = dbus.pydbusconn.registerWithCaptcha
            opts.append(body.captcha)
        else:
            method = dbus.pydbusconn.register
        result = method(*opts)
        # successful verification just returns None
        if result:
            logger.info(result)
            return json({"error": result}, 400)
        return json(None, 200)
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, 'message', repr(err))
        logger.error(error)
        return json({"error": error}, 400)


@dataclass
class VerifyV1PostParams:
    """
    VerifyV1PostParams
    """

    pin: Optional[str] = field(default_factory=str)


@verify_v1.middleware("request")
async def verify_post_dummy_body(request):
    """
    Create empty dummy body for @validate
    if there is none in request
    """
    if not request.body:
        request.body = b"{}"


@verify_v1.post("/<number:path>/verify/<token:path>", version=1)
@openapi.tag("Devices")
@openapi.parameter("number", str, required=True, location="path")
@openapi.parameter("token", str, required=True, location="path")
@openapi.body({"application/json": VerifyV1PostParams})
@openapi.response(201, {"application/json": None}, description="Created")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Verify a registered phone number with the signal network.")
@validate(VerifyV1PostParams)
async def verify_post(
    request, number, token, body: VerifyV1PostParams
):  # pylint: disable=unused-argument
    """
    Verify a registered phone number.
    """
    opts = [number, token]
    try:
        dbus = SignalCLIDBus()
        if body.pin:
            method = dbus.pydbusconn.verifyWithPin
            opts.append(body.pin)
        else:
            method = dbus.pydbusconn.verify
        result = method(*opts)
        # successful verification just returns None
        if result:
            logger.info(result)
            return json({"error": result}, 400)
        return json(None, 200)
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, 'message', repr(err))
        logger.error(error)
        return json({"error": error}, 400)



"""
REST Wrapper for signal-cli running on system dbus
"""


from sanic import Sanic

from signal_cli_dbus_rest_api.api import entrypoint
from signal_cli_dbus_rest_api.version import __version__

app = Sanic("signal-cli-dbus-rest-api", env_prefix="SIGNAL_CLI_DBUS_REST_API_")
app.config.FALLBACK_ERROR_FORMAT = "json"
app.config.API_HOST = (
    f'{app.config.get("API_HOST", "127.0.0.1")}:{app.config.get("API_PORT", 8080)}'
)
app.config.API_BASEPATH = app.config.get("API_BASEPATH")
app.config.API_SCHEMES = app.config.get("API_SCHEMES", "http,https").split(",")
app.config.API_LICENSE_NAME = "MIT"
app.config.API_LICENSE_URL = "https://mit-license.org/"
app.config.API_TITLE = "Signal Cli REST API"
app.config.API_DESCRIPTION = "This is the Signal Cli REST API documentation."
app.config.API_VERSION = __version__
app.blueprint(entrypoint)


def run():
    """
    run app
    """
    app.run(
        host=app.config.get("HOST", "127.0.0.1"),
        port=app.config.get("PORT", 8080),
        debug=app.config.get("DEBUG", False),
        workers=app.config.get("WORKERS", 1),
        access_log=app.config.get("ACCESS_LOG", False),
        auto_reload=app.config.get("AUTO_RELOAD", False),
    )



"""
DBus connection for signal-cli
"""

from re import sub

from pydbus import SystemBus as pydbusSystemBus

from dbus import SystemBus as dbusSystemBus


class SignalCLIDBus:  # pylint: disable=too-few-public-methods
    """
    DBus connection for signal-cli
    """

    def __init__(self, **kwargs):
        """
        initialize DBus connection
        """
        self.connection = None
        self.pydbus = pydbusSystemBus()
        self.dbus = dbusSystemBus()
        objectpath = None
        if kwargs.get("number"):
            number_escape = (
                str(kwargs.get("number")).encode().decode("unicode_escape").strip("+")
            )
            objectpath = f"/org/asamk/Signal/_{number_escape}"
        if objectpath and kwargs.get("groupid"):
            groupid = sub(r"[+|=|/]", "_", kwargs.get("groupid"))
            objectpath += f"/Groups/{groupid}"
        if kwargs.get("account"):
            objectpath = kwargs.get("account")
        if objectpath:
            self.pydbusconn = self.pydbus.get(
                "org.asamk.Signal",
                objectpath,
            )
            self.dbusconn = self.dbus.get_object(
                "org.asamk.Signal",
                objectpath,
            )
        else:
            self.pydbusconn = self.pydbus.get("org.asamk.Signal")
            self.dbusconn = self.dbus.get_object(
                "org.asamk.Signal", "/org/asamk/Signal"
            )



"""
send messages to single contact or group
"""


from dataclasses import dataclass, field
from os import remove as os_remove
from typing import List, Optional
from uuid import uuid4

from sanic import Blueprint, Sanic
from sanic.log import logger
from sanic.response import json
from sanic_ext import openapi, validate
from signal_cli_dbus_rest_api.dataclasses import Error, ResponseTimestamp
from signal_cli_dbus_rest_api.lib.dbus import SignalCLIDBus
from signal_cli_dbus_rest_api.lib.helper import (get_groupid_as_bytes,
                                                 is_phone_number,
                                                 do_decode_attachments)

send_v1 = Blueprint("send_v1", url_prefix="/send")
send_v2 = Blueprint("send_v2", url_prefix="/send")


def do_send_message(
    recipients: list, number: str, message: str, attachments: list, version: int = 2
):
    """
    send message
    """
    try:
        dbus = SignalCLIDBus(number=number)
        real_recipients = is_phone_number(
            recipients=recipients, dbus_connection=dbus.dbusconn
        )
        method = dbus.dbusconn.sendMessage
        signature = "sasas"
        if not real_recipients:
            real_recipients = get_groupid_as_bytes(
                recipient=recipients[0], version=version
            )
            method = dbus.dbusconn.sendGroupMessage
            signature = "sasay"
        if not real_recipients:
            return (
                {
                    "error": f"{recipients} is neither a phone number nor a valid group id"
                },
                400,
            )
        timestamp = method(message, attachments, real_recipients, signature=signature)
        return {"timestamp": str(timestamp)}, 201
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, 'message', repr(err))
        logger.error(error)
        return {"error": error}, 400
    finally:
        for attachment in attachments:
            os_remove(attachment)


@dataclass
class SendV2PostParams:
    """
    SendV2PostParams
    """

    recipients: List[str]
    message: str
    number: Optional[str] = field(default_factory=str)
    base64_attachments: Optional[List[str]] = field(default_factory=list)


@send_v2.post("/", version=2)
@openapi.tag("Messages")
@openapi.body({"application/json": SendV2PostParams}, required=True)
@openapi.response(201, {"application/json": ResponseTimestamp}, description="Created")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description(
    (
        "Send a signal message."
        "`number` can be ommited if API is running w/ `SIGNAL_CLI_DBUS_REST_API_ACCOUNT`"
    )
)
@validate(SendV2PostParams)
async def send_v2_post(request, body: SendV2PostParams):  # pylint: disable=unused-argument
    """
    Send a signal message.
    """
    decoded_attachments = []
    app = Sanic.get_app()
    recipients = body.recipients
    try:
        number = body.number or app.config.ACCOUNT
    except AttributeError:
        return json(
            {
                "error": "number missing in request and SIGNAL_CLI_DBUS_REST_API_ACCOUNT unset "
            },
            400,
        )
    attachments = body.base64_attachments
    uuid = str(uuid4())
    if isinstance(attachments, list):
        decoded_attachments = do_decode_attachments(attachments, uuid)
    return_message, return_code = do_send_message(
        recipients, number, body.message, decoded_attachments
    )
    return json(return_message, return_code)


@dataclass
class SendV1PostParams:
    """
    SendV1PostParams
    """

    message: str
    number: Optional[str] = field(default_factory=str)
    base64_attachments: List[str] = field(default_factory=list)


@send_v1.post("/<recipient:path>", version=1)
@openapi.tag("Messages")
@openapi.parameter("recipient", str, required=True, location="path")
@openapi.body({"application/json": SendV1PostParams}, required=True)
@openapi.response(201, {"application/json": ResponseTimestamp}, description="Created")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description(
    (
        "Send a signal message."
        "`number` can be ommited if API is running w/ `SIGNAL_CLI_DBUS_REST_API_ACCOUNT`"
    )
)
@validate(SendV1PostParams)
async def send_v1_post(request, recipient, body: SendV1PostParams):  # pylint: disable=unused-argument
    """
    Send a signal message.
    """
    decoded_attachments = []
    app = Sanic.get_app()
    try:
        number = body.number or app.config.ACCOUNT
    except AttributeError:
        return json(
            {
                "error": "number missing in request and SIGNAL_CLI_DBUS_REST_API_ACCOUNT unset "
            },
            400,
        )
    attachments = body.base64_attachments
    uuid = str(uuid4())
    if isinstance(attachments, list):
        decoded_attachments = do_decode_attachments(attachments, uuid)
    return_message, return_code = do_send_message(
        [recipient], number, body.message, decoded_attachments, version=1
    )
    return json(return_message, return_code)



"""
search handler
"""

from typing import List
from dataclasses import dataclass
from sanic import Blueprint
from sanic.log import logger
from sanic.response import json
from sanic_ext import openapi, validate
from signal_cli_dbus_rest_api.lib.dbus import SignalCLIDBus
from signal_cli_dbus_rest_api.dataclasses import Error

search_v1 = Blueprint("search_v1", url_prefix="/search")


@dataclass
class SearchV1GetParams:
    """
    SearchV1GetParams
    """

    numbers: List[str]


@dataclass
class SearchV1GetResponse:
    """
    SearchV1GetResponse
    """

    number: str
    registered: bool


@search_v1.get("/", version=1)
@openapi.tag("Search")
@openapi.parameter(
    "numbers",
    List[str],
    required=True,
    location="query",
    description="Numbers to check",
)
@openapi.response(
    200,
    {
        "application/json": List[SearchV1GetResponse],
    },
    description="OK",
)
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description(
    "Check if one or more phone numbers are registered with the Signal Service."
)
@validate(query=SearchV1GetParams)
async def search_v1_get(
    request, query: SearchV1GetParams
):  # pylint: disable=unused-argument
    """
    Check if one or more phone numbers are registered with the Signal Service.
    """
    numbers = request.args.getlist("numbers")
    try:
        dbus = SignalCLIDBus()
        accounts = dbus.pydbusconn.listAccounts()
        network_result = []
        result = []
        for account in accounts:
            account_dbus = SignalCLIDBus(account=account)
            network_result = account_dbus.dbusconn.isRegistered(numbers, signature="as")
            if network_result:
                break
        data = dict(zip(numbers, network_result))
        for key, value in data.items():
            result.append({"number": key, "registered": bool(value)})
        return json(result, 200)
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, 'message', repr(err))
        logger.error(error)
        return json({"error": error}, 400)



"""
about
"""

from typing import List
from sanic import Blueprint
from sanic.response import json
from sanic_ext import openapi

about_v1 = Blueprint("about_v1", url_prefix="/about")


@about_v1.get("/", version=1)
@openapi.tag("General")
@openapi.response(
    200,
    {
        "application/json": {
            "mode": str,
            "versions": List[str],
        }
    },
    description="OK",
)
@openapi.description("Returns the supported API versions.")
async def about_v1_get(request):  # pylint: disable=unused-argument
    """
    Lists general information about the API.
    """
    return json(
        {
            "mode": "dbus",
            "versions": ["v1", "v2"],
        },
        200
    )



"""
helpers
"""

from base64 import b64decode, b64encode
from binascii import Error as binascii_Error
from io import BytesIO
from mimetypes import guess_extension
from re import sub
from tempfile import mkstemp

from magic import from_buffer
from sanic.log import logger


def get_groupid_as_bytes(recipient: str, version: int = 2):
    """
    convert base64 encoded group id to bytes
    see https://github.com/AsamK/signal-cli/issues/272
    and https://github.com/AsamK/signal-cli/wiki/DBus-service
    """
    if version == 1:
        recipient = recipient.replace("_", "/")
    try:
        return b64decode(
            recipient, validate=True
        )
    except binascii_Error:
        return None


def is_phone_number(recipients: list, dbus_connection: object):
    """
    check if recipient is a valid phone number
    """
    if not isinstance(recipients, list):
        return None
    if isinstance(recipients[0], bytes):
        return None
    for recipient in recipients:
        if not (recipient.startswith("0") or recipient.startswith("+")):
            recipient = f"+{recipient}"
        if not dbus_connection.isRegistered(recipient):
            return None
    return recipients


def get_group_properties(systembus: object, number: str = "",
                         objectpath: str = "", groupid: str = ""):
    """
    get group properties
    """
    if not objectpath:
        if not number and groupid:
            return (False, "Missing number and groupid")
        objectpath = (
            f"/org/asamk/Signal/_{number.strip('+')}/"
            f"Groups/{sub(r'[+|=|/]', '_', groupid)}"
        )
    try:
        result = systembus.get(
            "org.asamk.Signal",
            objectpath,
        ).GetAll("org.asamk.Signal.Group")
        return (
            True,
            {
                "blocked": result.get("IsBlocked"),
                "id": b64encode(bytearray(result.get("Id"))).decode(),
                "internal_id": result.get("Id"),
                "invite_link": result.get("GroupInviteLink"),
                "members": result.get("Members"),
                "name": result.get("Name"),
                "pending_invites": result.get("PendingMembers"),
                "pending_requests": result.get("RequestingMembers"),
                "message_expiration_timer": result.get("MessageExpirationTimer"),
                "admins": result.get("Admins"),
                "description": result.get("Description"),
            }
        )
    # pylint: disable=broad-except
    except Exception as err:
        return (False, err)


def do_decode_attachments(attachments, uuid):
    """
    decode base64 attachments and dump the decoded
    content to disk for sending out later
    """
    decoded_attachments = []
    for index, attachment in enumerate(attachments):
        try:
            attachment_io_bytes = BytesIO()
            attachment_io_bytes.write(b64decode(attachment))
            extension = guess_extension(
                from_buffer(attachment_io_bytes.getvalue(), mime=True)
            )
            _, filename = mkstemp(prefix=f"{uuid}_{index}_", suffix=f".{extension}")
            with open(filename, "wb") as f_h:
                f_h.write(b64decode(attachment))
            decoded_attachments.append(filename)
        # pylint: disable=broad-except
        except Exception as err:
            logger.error("unable to decode attachment: %s", err)
    return decoded_attachments



"""
signal-cli-dbus-rest-api
"""

import codecs
import os.path
from pathlib import Path

from setuptools import find_packages, setup


def read(rel_path):
    """
    read
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as f_p:
        return f_p.read()


def get_version(rel_path):
    """
    get_version
    """
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


# read the contents of your README file
DIR = Path(__file__).parent
LONG_DESCRIPTION = (DIR / "README.md").read_text()

setup(
    name="signal-cli-dbus-rest-api",
    author="Stefan Heitmüller",
    author_email="stefan.heitmueller@gmx.com",
    license="MIT",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    python_requires="~=3.7",
    packages=find_packages(),
    version=get_version("signal_cli_dbus_rest_api/version/__init__.py"),
    install_requires=[
        "sanic==21.12.1",
        "sanic-ext==22.1.2",
        "python-magic==0.4.24",
        "pydbus==0.6.0",
    ],
    entry_points={
        "console_scripts": ["signal-cli-dbus-rest-api=signal_cli_dbus_rest_api:run"],
    },
)



"""
groups handler
"""

from base64 import b64encode
from dataclasses import dataclass, field
from enum import Enum
from os import remove as os_remove
from typing import Dict, List, Optional
from uuid import uuid4

from sanic import Blueprint, Sanic
from sanic.log import logger
from sanic.response import json, empty
from sanic_ext import openapi, validate
from signal_cli_dbus_rest_api.dataclasses import Error, GroupId
from signal_cli_dbus_rest_api.lib.dbus import SignalCLIDBus
from signal_cli_dbus_rest_api.lib.helper import (do_decode_attachments,
                                                 get_group_properties)

create_group_v1 = Blueprint("create_group_v1", url_prefix="/groups")
delete_group_v1 = Blueprint("delete_group_v1", url_prefix="/groups")
groups_for_number_v1 = Blueprint("groups_of_number_v1", url_prefix="/groups")
group_details_v1 = Blueprint("group_details_v1", url_prefix="/groups")
join_group_v1 = Blueprint("join_group_v1", url_prefix="/groups")
quit_group_v1 = Blueprint("quit_group_v1", url_prefix="/groups")
update_group_v1 = Blueprint("update_group_v1", url_prefix="/groups")


@dataclass
class GroupsForNumberGetV1Params:
    """
    GroupsForNumberGetV1Params
    """

    number: str


@dataclass
class GroupsForNumberGetV1ResponseItem:  # pylint: disable=too-many-instance-attributes
    """
    GroupsForNumberGetV1ResponseItem
    """

    blocked: bool
    id: str  # pylint: disable=invalid-name
    internal_id: str
    invite_link: str
    members: List[str]
    name: str
    pending_invites: List[str]
    pending_requests: List[str]
    message_expiration_timer: int
    admins: List[str]
    description: str


@groups_for_number_v1.get("/<number:path>", version=1)
@openapi.tag("Groups")
@openapi.parameter(
    "number",
    str,
    required=True,
    location="path",
    description="Registered Phone Number",
)
@openapi.response(
    200,
    {"application/json": List[GroupsForNumberGetV1ResponseItem]},
    description="OK",
)
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("List all Signal Groups.")
async def groups_for_number_get(request, number):  # pylint: disable=unused-argument
    """
    List all Signal Groups.
    """
    try:
        dbus = SignalCLIDBus(number=number)
        groups = dbus.pydbusconn.listGroups()
        result = []
        for group in groups:
            success, data = get_group_properties(
                systembus=dbus.pydbus,
                objectpath=group[0],
            )
            if not success:
                return json({"error": data}, 400)
            result.append(data)
        return json(result, 200)
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return json({"error": error}, 400)


@group_details_v1.get("/<number:path>/<groupid:path>", version=1)
@openapi.tag("Groups")
@openapi.parameter(
    "number",
    str,
    required=True,
    location="path",
    description="Registered Phone Number",
)
@openapi.parameter(
    "groupid",
    str,
    required=True,
    location="path",
    description="Group ID",
)
@openapi.response(
    200,
    {"application/json": GroupsForNumberGetV1ResponseItem},
    description="OK",
)
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("List a Signal Group.")
async def groups_of_number_get(
        request, number: str, groupid: str
):  # pylint: disable=unused-argument
    """
    List a Signal Group.
    """
    try:
        dbus = SignalCLIDBus()
        success, data = get_group_properties(
            systembus=dbus.pydbus,
            number=number,
            groupid=groupid,
        )
        if not success:
            return json({"error": data}, 400)
        return json(data, 200)
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return json({"error": error}, 400)


@dataclass
class CreateGroupV1Permissions:
    """
    CreateGroupV1Permissions
    """

    add_members: str = "only-admins"
    edit_group: str = "only-admins"


@dataclass
class GroupLinkV1Choices(Enum):
    """
    GroupLinkV1Choices
    """

    DISABLED = "disabled"
    ENABLED = "enabled"
    ENABLED_WITH_APPROVAL = "enabled-with-approval"


def update_group_link(pydbusconn, setting):
    """
    return proper method to call for group link handling
    """
    # pylint: disable=no-member
    approvals = {
        GroupLinkV1Choices.ENABLED.value: False,
        GroupLinkV1Choices.ENABLED_WITH_APPROVAL.value: True,
    }
    methods = {
        GroupLinkV1Choices.DISABLED.value: pydbusconn.disableLink,
        GroupLinkV1Choices.ENABLED.value: pydbusconn.enableLink,
        GroupLinkV1Choices.ENABLED_WITH_APPROVAL.value: pydbusconn.enableLink,
    }
    method = methods.get(setting, None)
    if method is None:
        return
    approval = approvals.get(setting, None)
    if approval is None:
        method()
        return
    method(approval)
    return


@dataclass
class CreateGroupV1PostParamsDocs:  # pylint: disable=too-many-instance-attributes
    """
    CreateGroupV1PostParams
    """

    name: str
    members: List[str]
    permissions: Optional[CreateGroupV1Permissions]
    group_link: Optional[GroupLinkV1Choices]
    admins: Optional[List[str]] = field(default_factory=list)
    description: Optional[str] = field(default_factory=str)
    base64_avatar: Optional[str] = field(default_factory=str)
    message_expiration_timer: Optional[int] = 0


@dataclass
class CreateGroupV1PostParamsValidate(CreateGroupV1PostParamsDocs):
    """
    CreateGroupV1PostParamsValidate
    """

    group_link: str = "disabled"
    permissions: Optional[Dict[str, CreateGroupV1Permissions]] = field(
        default_factory=dict
    )

    def __post_init__(self):
        if isinstance(self.permissions, dict):
            self.permissions = CreateGroupV1Permissions(**self.permissions)


@create_group_v1.post("/<number:path>", version=1)
@openapi.tag("Groups")
@openapi.parameter("number", str, required=True, location="path")
@openapi.body({"application/json": CreateGroupV1PostParamsDocs}, required=True)
@openapi.response(201, {"application/json": GroupId}, description="Created")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Create a new Signal Group.")
@validate(CreateGroupV1PostParamsValidate)
async def create_group_v1_post(
    request, number, body: CreateGroupV1PostParamsValidate
):  # pylint: disable=unused-argument
    """
    Create a new Signal Group with the specified members.
    """
    avatar = ""
    app = Sanic.get_app()
    try:
        number = number or app.config.ACCOUNT
    except AttributeError:
        return json(
            {
                "error": "number missing in request and SIGNAL_CLI_DBUS_REST_API_ACCOUNT unset "
            },
            400,
        )
    uuid = str(uuid4())
    try:
        avatar = do_decode_attachments([body.base64_avatar], uuid)[0]
    # pylint: disable=broad-except
    except IndexError:
        pass
    try:
        dbus = SignalCLIDBus(number=number)
        groupid = dbus.pydbusconn.createGroup(
            body.name,
            body.members,
            avatar,
        )
        b64_groupid = b64encode(bytearray(groupid)).decode()
        dbus = SignalCLIDBus(
            number=number,
            groupid=b64_groupid
        )
        update_group_link(dbus.pydbusconn, body.group_link)
        dbus.pydbusconn.Description = body.description
        dbus.pydbusconn.MessageExpirationTimer = body.message_expiration_timer
        dbus.pydbusconn.PermissionAddMember = (
            body.permissions.add_members.upper().replace("-", "_")
        )
        dbus.pydbusconn.PermissionEditDetails = (
            body.permissions.edit_group.upper().replace("-", "_")
        )
        if body.admins:
            dbus.pydbusconn.addAdmins(body.admins)
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return json({"error": error}, 400)
    finally:
        os_remove(avatar)
    return json({"id": b64_groupid}, 201)


@delete_group_v1.delete("/<number:path>/<groupid:path>", version=1)
@openapi.tag("Groups")
@openapi.parameter("number", str, required=True, location="path")
@openapi.parameter("groupid", str, required=True, location="path")
@openapi.response(204, None, description="OK")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Delete a Signal Group.")
async def delete_group_v1_delete(
    request, number: str, groupid: str
):  # pylint: disable=unused-argument
    """
    Delete the specified Signal Group.
    """
    try:
        dbus = SignalCLIDBus(number=number, groupid=groupid)
        dbus.pydbusconn.deleteGroup()
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return json({"error": error}, 400)
    return empty


@dataclass
class UpdateGroupLinkV1Choices(Enum):
    """
    UpdateGroupLinkV1Choices
    """

    UNCHANGED = "unchanged"
    DISABLED = "disabled"
    ENABLED = "enabled"
    ENABLED_WITH_APPROVAL = "enabled-with-approval"


@dataclass
class UpdateGroupV1PatchParamsDocs:
    """
    UpdateGroupV1PatchParamsDocs
    """

    group_link: Optional[UpdateGroupLinkV1Choices]
    add_admins: Optional[List[str]] = field(default_factory=list)
    add_members: Optional[List[str]] = field(default_factory=list)


@dataclass
class UpdateGroupV1PatchParamsValidate(UpdateGroupV1PatchParamsDocs):
    """
    UpdateGroupV1PatchParamsValidate
    """

    group_link: str = "unchanged"


@update_group_v1.patch("/<number:path>/<groupid:path>", version=1)
@openapi.tag("Groups")
@openapi.parameter("number", str, required=True, location="path")
@openapi.parameter("groupid", str, required=True, location="path")
@openapi.body({"application/json": UpdateGroupV1PatchParamsDocs})
@openapi.response(
    200,
    {"application/json": GroupsForNumberGetV1ResponseItem},
    description="OK",
)
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Update an existing Signal Group.")
@validate(UpdateGroupV1PatchParamsValidate)
async def update_group_v1_patch(
        request, number: str, groupid: str, body: UpdateGroupV1PatchParamsValidate
):  # pylint: disable=unused-argument
    """
    Update an existing Signal Group.
    """
    try:
        dbus = SignalCLIDBus(number=number, groupid=groupid)
        if body.add_admins:
            dbus.pydbusconn.addAdmins(body.add_admins)
        if body.add_members:
            dbus.pydbusconn.addMembers(body.add_members)
        update_group_link(dbus.pydbusconn, body.group_link)
        dbus = SignalCLIDBus()
        success, data = get_group_properties(
            systembus=dbus.pydbus,
            number=number,
            groupid=groupid,
        )
        if not success:
            return json({"error": data}, 400)
        return json(data, 200)
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return json({"error": error}, 400)
    return json(data, 200)


@quit_group_v1.post("/<number:path>/<groupid:path>/quit", version=1)
@openapi.tag("Groups")
@openapi.parameter("number", str, required=True, location="path")
@openapi.parameter("groupid", str, required=True, location="path")
@openapi.body({"application/json": UpdateGroupV1PatchParamsDocs})
@openapi.response(204, None, description="OK")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Quit (leave) a Signal Group.")
async def quit_group_v1_post(
        request, number: str, groupid: str
):  # pylint: disable=unused-argument
    """
    Quit (leave) a Signal Group.
    """
    try:
        dbus = SignalCLIDBus(number=number, groupid=groupid)
        dbus.pydbusconn.quitGroup()
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return json({"error": error}, 400)
    return empty


@join_group_v1.post("/<number:path>/<groupid:path>/join", version=1)
@openapi.tag("Groups")
@openapi.parameter("number", str, required=True, location="path")
@openapi.parameter(
    "groupid",
    str,
    required=True,
    location="path",
    description="Get from invite link: https://signal.group/#`${groupid}`")
@openapi.response(204, None, description="OK")
@openapi.response(400, {"application/json": Error}, description="Bad Request")
@openapi.description("Join a Signal Group.")
async def join_group_v1_post(
        request, number: str, groupid: str
):  # pylint: disable=unused-argument
    """
    Join a Signal Group.
    """
    try:
        dbus = SignalCLIDBus(number=number)
        dbus.pydbusconn.joinGroup(f"https://signal.group/#{groupid}")
    # pylint: disable=broad-except
    except Exception as err:
        error = getattr(err, "message", repr(err))
        logger.error(error)
        return json({"error": error}, 400)
    return empty
